#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typing
from typing import Any, Dict, List, Union

from ..libs.alibabacloud_aiworkspace20210204.models import (
    CreateMemberRequest,
    CreateMemberRequestMembers,
    CreateMemberResponseBody,
    CreateWorkspaceRequest,
    CreateWorkspaceResponseBody,
    DeleteMembersRequest,
    GetDefaultWorkspaceRequest,
    GetDefaultWorkspaceResponseBody,
    GetMemberRequest,
    GetMemberResponseBody,
    GetWorkspaceRequest,
    GetWorkspaceResponseBody,
    ListConfigsRequest,
    ListConfigsResponseBody,
    ListMembersRequest,
    ListMembersResponseBody,
    ListProductAuthorizationsRequest,
    ListProductAuthorizationsResponseBody,
    ListWorkspacesRequest,
    ListWorkspacesResponseBody,
    UpdateConfigsRequest,
    UpdateConfigsRequestConfigs,
)
from .base import ResourceAPI, ServiceName

if typing.TYPE_CHECKING:
    pass


class WorkspaceConfigKeys(object):

    DEFAULT_OSS_STORAGE_URI = "modelExportPath"
    MAXCOMPUTE_TEMP_TABLE_LIFECYCLE = "tempTableLifecycle"


class WorkspaceAPI(ResourceAPI):

    BACKEND_SERVICE_NAME = ServiceName.PAI_WORKSPACE

    _list_method = "list_workspaces_with_options"
    _get_method = "get_workspace_with_options"
    _create_method = "create_workspace_with_options"
    _get_default_workspace_method = "get_default_workspace_with_options"
    _list_configs_method = "list_configs_with_options"
    _update_configs_method = "update_configs_with_options"

    _list_member_method = "list_members_with_options"
    _create_member_method = "create_member_with_options"
    _get_member_method = "get_member_with_options"
    _delete_members_method = "delete_members_with_options"
    _add_member_role_method = "add_member_role_with_options"
    _remove_member_role_method = "remove_member_role_with_options"
    _list_product_authorizations_method = "list_product_authorizations_with_options"

    def list(
        self,
        page_number=None,
        page_size=None,
        sort_by=None,
        order=None,
        name=None,
        module_list=None,
        status=None,
        option=None,
        verbose=None,
    ) -> List[Dict[str, Any]]:
        request = ListWorkspacesRequest(
            page_number=page_number,
            page_size=page_size,
            sort_by=sort_by,
            order=order,
            workspace_name=name,
            module_list=module_list,
            status=status,
            option=option,
            verbose=verbose,
        )
        res: ListWorkspacesResponseBody = self._do_request(
            method_=self._list_method, request=request
        )

        return [item.to_map() for item in res.workspaces]

    def get(self, workspace_id: str, verbose: bool = True) -> Dict[str, Any]:
        request = GetWorkspaceRequest(verbose=verbose)

        res: GetWorkspaceResponseBody = self._do_request(
            method_=self._get_method,
            workspace_id=workspace_id,
            request=request,
        )
        return res.to_map()

    def create(
        self,
        name: str,
        display_name: str = None,
        description: str = None,
        env_types: List[str] = None,
    ) -> str:
        request = CreateWorkspaceRequest(
            description=description,
            display_name=display_name,
            workspace_name=name,
            env_types=env_types,
        )

        res: CreateWorkspaceResponseBody = self._do_request(
            method_=self._create_method, request=request
        )
        return res.workspace_id

    def list_members(
        self,
        workspace_id: str,
        member_name: str = None,
        roles: List[str] = None,
        page_number: int = None,
        page_size: int = None,
    ) -> List[Dict[str, Any]]:
        request = ListMembersRequest(
            member_name=member_name,
            page_number=page_number,
            page_size=page_size,
            roles=roles,
        )

        res: ListMembersResponseBody = self._do_request(
            method_=self._list_member_method,
            workspace_id=workspace_id,
            request=request,
        )
        return [item.to_map() for item in res.members]

    def add_member(self, workspace_id: str, user_id: str, roles: List[str]) -> str:
        request = CreateMemberRequest(
            members=[
                CreateMemberRequestMembers(
                    user_id=user_id,
                    roles=roles,
                )
            ],
        )

        res: CreateMemberResponseBody = self._do_request(
            method_=self._create_member_method,
            workspace_id=workspace_id,
            request=request,
        )
        return res.members[0].member_id

    def get_member(self, workspace_id: str, user_id: str) -> Dict[str, Any]:
        request = GetMemberRequest(user_id=user_id)

        res: GetMemberResponseBody = self._do_request(
            method_=self._get_member_method, workspace_id=workspace_id, request=request
        )
        return res.to_map()

    def delete_members(self, workspace_id: str, member_ids: List[str]) -> None:
        request = DeleteMembersRequest(
            member_ids=member_ids,
        )
        self._do_request(
            method_=self._delete_members_method,
            workspace_id=workspace_id,
            request=request,
        )

    def add_member_role(
        self, workspace_id: str, member_id: str, role_name: str
    ) -> None:
        self._do_request(
            method_=self._add_member_role_method,
            workspace_id=workspace_id,
            member_id=member_id,
            role_name=role_name,
        )

    def remove_member_role(
        self, workspace_id: str, member_id: str, role_name: str
    ) -> None:
        self._do_request(
            method_=self._remove_member_role_method,
            workspace_id=workspace_id,
            member_id=member_id,
            role_name=role_name,
        )

    def get_default_workspace(self) -> Dict[str, Any]:
        """Get the default workspace for the account."""
        request = GetDefaultWorkspaceRequest()
        resp: GetDefaultWorkspaceResponseBody = self._do_request(
            method_=self._get_default_workspace_method, request=request
        )
        return resp.to_map()

    def list_configs(self, workspace_id, config_keys: Union[List[str], str]) -> Dict:
        """List configs used in the Workspace."""
        request = ListConfigsRequest(
            config_keys=(
                ",".join(config_keys)
                if isinstance(config_keys, (tuple, list))
                else config_keys
            ),
        )

        resp: ListConfigsResponseBody = self._do_request(
            method_=self._list_configs_method,
            workspace_id=workspace_id,
            request=request,
        )
        return resp.to_map()

    def get_default_storage_uri(self, workspace_id):
        resp = self.list_configs(
            workspace_id=workspace_id,
            config_keys=WorkspaceConfigKeys.DEFAULT_OSS_STORAGE_URI,
        )

        oss_storage_uri = next(
            (
                item["ConfigValue"]
                for item in resp["Configs"]
                if item["ConfigKey"] == WorkspaceConfigKeys.DEFAULT_OSS_STORAGE_URI
            ),
            None,
        )
        return oss_storage_uri

    def update_configs(self, workspace_id: str, configs: Union[Dict, List]):
        """Update configs used in the Workspace."""
        if isinstance(configs, Dict):
            configs = [
                UpdateConfigsRequestConfigs(
                    config_key=key,
                    config_value=value,
                )
                for key, value in configs.items()
            ]
        else:
            configs = [UpdateConfigsRequestConfigs.from_map(item) for item in configs]
        request = UpdateConfigsRequest(configs=configs)

        self._do_request(
            method_=self._update_configs_method,
            workspace_id=workspace_id,
            request=request,
        )

    def list_product_authorizations(self, ram_role_names: List[str]) -> Dict[str, Any]:
        request = ListProductAuthorizationsRequest(
            ram_role_names=",".join(ram_role_names)
        )

        res: ListProductAuthorizationsResponseBody = self._do_request(
            self._list_product_authorizations_method, request
        )

        return res.to_map()



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import posixpath
import time
from typing import Optional

import requests

from .common.oss_utils import is_oss_uri
from .exception import UnexpectedStatusException
from .session import Session, get_default_session


class TensorBoardStatus(object):
    Pending = "Pending"
    Creating = "Creating"
    Running = "Running"
    Creating_Failed = "Creating_Failed"
    Deleting = "Deleting"
    Deleted = "Deleted"
    Stopping = "Stopping"
    Stopped = "Stopped"

    @classmethod
    def is_terminated(cls, status):
        return status in [
            cls.Creating_Failed,
            cls.Stopped,
            cls.Deleted,
        ]

    @classmethod
    def is_running(cls, status):
        return status in [
            cls.Running,
        ]


class TensorBoard(object):
    def __init__(self, tensorboard_id: str, session: Optional[Session] = None):
        self.session = session or get_default_session()
        self.tensorboard_id = tensorboard_id
        self._api_object = self.session.tensorboard_api.get(tensorboard_id)

    def __repr__(self):
        return "TensorBoard(tensorboard_id={}, name={}, status={})".format(
            self.tensorboard_id,
            self.display_name,
            self._status(),
        )

    @property
    def status(self):
        self._refresh()
        return self._status()

    def _status(self):
        return self._api_object.get("Status")

    @property
    def app_uri(self):
        """Get the TensorBoard application URI."""
        return self._api_object.get("TensorboardUrl")

    @property
    def summary_uri(self):
        return self._api_object.get("SummaryUri")

    @property
    def display_name(self):
        return self._api_object.get("DisplayName")

    def _refresh(self):
        self._api_object = self.session.tensorboard_api.get(self.tensorboard_id)

    @classmethod
    def create(
        cls,
        uri: str,
        wait: bool = True,
        display_name: Optional[str] = None,
        max_runtime_in_minutes: Optional[int] = None,
        source_id: Optional[str] = None,
        source_type: Optional[str] = None,
        session: Optional[Session] = None,
    ) -> "TensorBoard":
        """Launch a TensorBoard Application.

        Args:
            uri (str): A OSS URI to the directory containing the TensorBoard logs.
            wait (bool): Whether to wait for the TensorBoard application to be ready.
            display_name (str, optional): Display name of the TensorBoard application.
                Defaults to None.
            max_runtime_in_minutes: Maximum running time in minutes.
            source_type (str, optional): The type of the source object. Defaults to None.
            source_id (str, optional): The ID of the source object. Defaults to None.
            session: A Session object to use in interacting with PAI.
        Returns:
            TensorBoard: A TensorBoard object.

        Examples:
            Create a TensorBoard application from a OSS URI:
            >>> from pai.tensorboard import TensorBoard
            >>> tb = TensorBoard.create("oss://my-bucket/path/to/logs_dir/")
            >>> # Get TensorBoard Application URL.
            >>> print(tb.app_uri)
        """
        session = session or get_default_session()

        if not is_oss_uri(uri):
            raise RuntimeError("Currently only support OSS uri to create TensorBoard.")
        oss_uri = session.patch_oss_endpoint(uri)
        data_source_type = "OSS"

        if not display_name:
            # Use the last part of the OSS URI as the display name.
            display_name = posixpath.basename(uri.rstrip("/"))
            if not display_name:
                raise RuntimeError("Failed to infer display name from OSS URI.")

        tb_id = session.tensorboard_api.create(
            uri=oss_uri,
            display_name=display_name,
            data_source_type=data_source_type,
            max_running_time_minutes=max_runtime_in_minutes,
            # hack: summary_relative_path is required for CreateTensorBoard API.
            summary_relative_path="/",
            source_id=source_id,
            source_type=source_type,
        )
        tensorboard = TensorBoard(tensorboard_id=tb_id, session=session)
        if wait:
            tensorboard.wait()
        return tensorboard

    def wait(self):
        """Wait for the TensorBoard application to be ready.

        Raises:
            UnExpectedStatusException: If the TensorBoard application is terminated
                unexpectedly.

        """
        while True:
            status = self.status
            if TensorBoardStatus.is_terminated(status):
                raise UnexpectedStatusException(
                    "TensorBoard terminated unexpectedly in status: %s" % status,
                    status,
                )
            elif TensorBoardStatus.is_running(status):
                self._wait_app_available()
                return
            else:
                time.sleep(5)
                self._refresh()

    def start(self, wait: bool = True):
        """Start the TensorBoard application."""
        self._refresh()
        if TensorBoardStatus.is_running(self.status):
            return

        self.session.tensorboard_api.start(self.tensorboard_id)
        if wait:
            self.wait()

    def stop(self):
        """Stop the TensorBoard application."""
        self._refresh()
        if not TensorBoardStatus.is_running(self.status):
            return

        self.session.tensorboard_api.stop(self.tensorboard_id)

    def _wait_app_available(self):
        """Wait until the TensorBoard application is available."""
        if not self.app_uri:
            raise RuntimeError("TensorBoard application URL is not available.")

        while True:
            resp = requests.get(self.app_uri)
            # Status code not equals 5xx means the TensorBoard application is available.
            if resp.status_code // 100 != 5:
                break
            time.sleep(5)

    def delete(self):
        """Delete the TensorBoard Application."""
        self.session.tensorboard_api.delete(self.tensorboard_id)



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List

from ..libs.alibabacloud_aiworkspace20210204.models import (
    CreateExperimentRequest,
    CreateExperimentResponseBody,
    Experiment,
    LabelInfo,
    ListExperimentRequest,
    ListExperimentResponseBody,
    SetExperimentLabelsRequest,
    UpdateExperimentRequest,
)
from .base import PaginatedResult, ServiceName, WorkspaceScopedResourceAPI


class ExperimentAPI(WorkspaceScopedResourceAPI):
    BACKEND_SERVICE_NAME = ServiceName.PAI_WORKSPACE

    _get_method = "get_experiment_with_options"
    _create_method = "create_experiment_with_options"
    _list_method = "list_experiment_with_options"
    _update_method = "update_experiment_with_options"
    _delete_method = "delete_experiment_with_options"
    _set_labels_method = "set_experiment_labels_with_options"
    _delete_label_method = "delete_experiment_label_with_options"

    def get(self, experiment_id: str):
        resp: Experiment = self._do_request(
            method_=self._get_method, experiment_id=experiment_id
        )
        return resp.to_map()

    def create(
        self,
        name,
        artifact_uri,
        workspace_id,
        **kwargs,
    ):
        request = CreateExperimentRequest(
            name=name,
            artifact_uri=artifact_uri,
            workspace_id=workspace_id,
            **kwargs,
        )
        resp: CreateExperimentResponseBody = self._do_request(
            method_=self._create_method, request=request
        )
        return resp.experiment_id

    def list(
        self,
        name: str = None,
        page_size: int = 50,
        page_number: int = 1,
        order: str = "DESC",
        **kwargs,
    ) -> PaginatedResult:
        workspace_id = kwargs.pop("workspace_id", None)
        request = ListExperimentRequest(
            name=name,
            page_size=page_size,
            page_number=page_number,
            order=order,
            workspace_id=workspace_id,
            **kwargs,
        )
        resp: ListExperimentResponseBody = self._do_request(
            method_=self._list_method, request=request
        )
        return self.make_paginated_result(resp)

    def update(
        self,
        experiment_id: str,
        name: str,
        **kwargs,
    ):
        request = UpdateExperimentRequest(
            name=name,
            **kwargs,
        )
        self._do_request(
            method_=self._update_method, experiment_id=experiment_id, request=request
        )
        return

    def delete(self, experiment_id: str):
        self._do_request(method_=self._delete_method, experiment_id=experiment_id)
        return

    def set_experiment_labels(self, experiment_id: str, labels: List[LabelInfo]):
        request = SetExperimentLabelsRequest(
            labels=labels,
        )
        self._do_request(
            method_=self._set_labels_method,
            experiment_id=experiment_id,
            request=request,
        )
        return

    def delete_experiment_label(self, experiment_id: str, key: str):
        self._do_request(
            method_=self._delete_label_method, experiment_id=experiment_id, key=key
        )
        return
